import math
import numpy as np
from coremltools.proto import FeatureTypes_pb2 as _FeatureTypes_pb2
from coremltools.proto import NeuralNetwork_pb2 as _NeuralNetwork_pb2


def _get_translator_function(layer_type):
    """Get the right translator function
    """
    if layer_type in _LAYER_REGISTERY:
        return _LAYER_REGISTERY[layer_type]
    else:
        raise TypeError(
            "Shape computation function missing for layer of type %s." % layer_type
        )


def _identity(layer, shape_dict):
    shape_dict[layer.output[0]] = shape_dict[layer.input[0]]


def _convolution(layer, shape_dict):
    params = layer.convolution
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]

    n_groups = params.nGroups
    Kh = Kw = 3
    hstride = wstride = hdilation = wdilation = 1
    if len(params.kernelSize) != 0:
        Kh, Kw = params.kernelSize
    if len(params.stride) != 0:
        hstride, wstride = params.stride
    if len(params.dilationFactor) != 0:
        hdilation, wdilation = params.dilationFactor
    Kh_dilated = (Kh - 1) * hdilation + 1
    Kw_dilated = (Kw - 1) * wdilation + 1
    l = r = b = t = 0
    if params.WhichOneof("ConvolutionPaddingType") == "valid":
        if len(params.valid.paddingAmounts.borderAmounts) != 0:
            t = params.valid.paddingAmounts.borderAmounts[0].startEdgeSize
            b = params.valid.paddingAmounts.borderAmounts[0].endEdgeSize
            l = params.valid.paddingAmounts.borderAmounts[1].startEdgeSize
            r = params.valid.paddingAmounts.borderAmounts[1].endEdgeSize
        if params.isDeconvolution:
            Hout = (Hin - 1) * hstride + Kh_dilated - t - b
            Wout = (Win - 1) * wstride + Kw_dilated - r - l
        else:
            Hout = (Hin + t + b - Kh_dilated) / hstride + 1
            Wout = (Win + r + l - Kw_dilated) / wstride + 1
    else:
        if params.isDeconvolution:
            Hout = Hin * hstride
            Wout = Win * wstride
        else:
            Hout = math.ceil(Hin / float(hstride))
            Wout = math.ceil(Win / float(wstride))

    if params.isDeconvolution:
        if len(params.outputShape) != 0:
            Hout, Wout = params.outputShape

    Cout = params.outputChannels
    shape_dict[layer.output[0]] = (Seq, Batch, int(Cout), int(Hout), int(Wout))


def _pooling(layer, shape_dict):
    params = layer.pooling
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]

    Kh = Kw = 3
    hstride = wstride = 1
    if len(params.kernelSize) != 0:
        Kh, Kw = params.kernelSize
    if len(params.stride) != 0:
        hstride, wstride = params.stride
    l = r = b = t = 0
    if params.globalPooling:
        Hout = Wout = 1
    else:
        if params.WhichOneof("PoolingPaddingType") == "valid":
            if len(params.valid.paddingAmounts.borderAmounts) != 0:
                t = params.valid.paddingAmounts.borderAmounts[0].startEdgeSize
                b = params.valid.paddingAmounts.borderAmounts[0].endEdgeSize
                l = params.valid.paddingAmounts.borderAmounts[1].startEdgeSize
                r = params.valid.paddingAmounts.borderAmounts[1].endEdgeSize
            Hout = (Hin + t + b - Kh) / hstride + 1
            Wout = (Win + r + l - Kw) / wstride + 1
        elif params.WhichOneof("PoolingPaddingType") == "same":
            Hout = math.ceil(Hin / float(hstride))
            Wout = math.ceil(Win / float(wstride))
        else:
            if len(params.includeLastPixel.paddingAmounts) != 0:
                t = params.includeLastPixel.paddingAmounts[0]
                b = t
                l = params.includeLastPixel.paddingAmounts[1]
                r = l
            Hout = math.ceil((Hin + 2 * t - Kh) / float(hstride)) + 1
            Wout = math.ceil((Win + 2 * l - Kw) / float(wstride)) + 1
            if t or l:
                if (Hout - 1) * hstride >= Hin + t:
                    Hout -= 1
                if (Wout - 1) * wstride >= Win + l:
                    Wout -= 1

    shape_dict[layer.output[0]] = (Seq, Batch, int(Cin), int(Hout), int(Wout))


def _inner_product(layer, shape_dict):
    params = layer.innerProduct
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]
    Cout = params.outputChannels
    shape_dict[layer.output[0]] = (Seq, Batch, int(Cout), 1, 1)


def _embedding(layer, shape_dict):
    params = layer.embedding
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]
    Cout = params.outputChannels
    shape_dict[layer.output[0]] = (Seq, Batch, int(Cout), 1, 1)


def _crop(layer, shape_dict):
    params = layer.crop
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]

    l = r = t = b = 0
    if len(layer.input) == 1:
        if len(params.cropAmounts.borderAmounts) != 0:
            t = params.cropAmounts.borderAmounts[0].startEdgeSize
            b = params.cropAmounts.borderAmounts[0].endEdgeSize
            l = params.cropAmounts.borderAmounts[1].startEdgeSize
            r = params.cropAmounts.borderAmounts[1].endEdgeSize
        Hout = Hin - t - b
        Wout = Win - l - r
    else:
        Hout = shape_dict[layer.input[1]][3]
        Wout = shape_dict[layer.input[1]][4]

    shape_dict[layer.output[0]] = (Seq, Batch, Cin, int(Hout), int(Wout))


def _padding(layer, shape_dict):
    params = layer.padding
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]

    l = r = t = b = 0
    if len(params.paddingAmounts.borderAmounts) != 0:
        t = params.paddingAmounts.borderAmounts[0].startEdgeSize
        b = params.paddingAmounts.borderAmounts[0].endEdgeSize
        l = params.paddingAmounts.borderAmounts[1].startEdgeSize
        r = params.paddingAmounts.borderAmounts[1].endEdgeSize
    Hout = Hin + t + b
    Wout = Win + l + r
    shape_dict[layer.output[0]] = (Seq, Batch, Cin, int(Hout), int(Wout))


def _upsample(layer, shape_dict):
    params = layer.upsample
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]

    sh = sw = 1
    if len(params.scalingFactor) != 0:
        sh, sw = params.scalingFactor
    Hout = Hin * sh
    Wout = Win * sw
    shape_dict[layer.output[0]] = (Seq, Batch, Cin, int(Hout), int(Wout))


def _add(layer, shape_dict):
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]

    for i, inp in enumerate(layer.input):
        if i == 0:
            continue
        _, _, c, h, w = shape_dict[inp]
        C = max(C, c)
        H = max(H, h)
        W = max(W, w)
    shape_dict[layer.output[0]] = (Seq, Batch, int(C), int(H), int(W))


def _dot(layer, shape_dict):
    Seq, Batch, _, _, _ = shape_dict[layer.input[0]]
    shape_dict[layer.output[0]] = (Seq, Batch, 1, 1, 1)


def _reduce(layer, shape_dict):
    params = layer.reduce
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]

    axis = _NeuralNetwork_pb2.ReduceLayerParams.ReduceAxis.Name(params.axis)
    if axis == "CHW":
        C = H = W = 1
    elif axis == "HW":
        H = W = 1
    elif axis == "C":
        C = 1
    elif axis == "H":
        H = 1
    elif axis == "W":
        W = 1

    shape_dict[layer.output[0]] = (Seq, Batch, int(C), int(H), int(W))


def _load_constant(layer, shape_dict):
    params = layer.loadConstant
    C, H, W = map(int, params.shape)
    shape_dict[layer.output[0]] = (1, 1, C, H, W)


def _reshape(layer, shape_dict):
    params = layer.reshape
    Seq, Batch, _, _, _ = shape_dict[layer.input[0]]

    if len(params.targetShape) == 3:
        C, H, W = params.targetShape
    else:
        Seq, C, H, W = params.targetShape

    shape_dict[layer.output[0]] = (int(Seq), Batch, int(C), int(H), int(W))


def _flatten(layer, shape_dict):
    params = layer.permute
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]

    shape_dict[layer.output[0]] = (int(Seq), int(Batch), int(Cin * Hin * Win), 1, 1)


def _permute(layer, shape_dict):
    params = layer.permute
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]

    axis = list(map(int, params.axis))
    dims = (Seq, Cin, Hin, Win)
    Seq_out = dims[axis[0]]
    Cout = dims[axis[1]]
    Hout = dims[axis[2]]
    Wout = dims[axis[3]]
    shape_dict[layer.output[0]] = (int(Seq_out), Batch, int(Cout), int(Hout), int(Wout))


def _concat(layer, shape_dict):
    params = layer.concat
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]

    if params.sequenceConcat:
        Seq = 0
        for inp in layer.input:
            Seq += shape_dict[inp][0]
    else:
        C = 0
        for inp in layer.input:
            C += shape_dict[inp][2]

    shape_dict[layer.output[0]] = (int(Seq), Batch, int(C), int(H), int(W))


def _split(layer, shape_dict):
    input_shape = shape_dict[layer.input[0]]
    Seq, Batch, C, H, W = input_shape
    for out in layer.output:
        shape_dict[out] = (Seq, Batch, C / len(layer.output), H, W)


def _sequence_repeat(layer, shape_dict):
    params = layer.sequenceRepeat
    n = params.nRepetitions
    if n == 0:
        n = 1
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]
    shape_dict[layer.output[0]] = (int(Seq * n), Batch, C, H, W)


def _reorganize_data(layer, shape_dict):
    params = layer.reorganizeData
    Seq, Batch, Cin, Hin, Win = shape_dict[layer.input[0]]
    block_size = params.blockSize
    Type = _NeuralNetwork_pb2.ReorganizeDataLayerParams.ReorganizationType.Name(
        params.mode
    )
    if Type == "SPACE_TO_DEPTH":
        Cout = Cin * block_size * block_size
        Hout = Hin / block_size
        Wout = Win / block_size
    else:
        Cout = Cin / (block_size * block_size)
        Hout = Hin * block_size
        Wout = Win * block_size
    shape_dict[layer.output[0]] = (Seq, Batch, int(Cout), int(Hout), int(Wout))


def _slice(layer, shape_dict):
    params = layer.slice
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]
    start = params.startIndex
    end = params.endIndex
    stride = params.stride
    axis = _NeuralNetwork_pb2.SliceLayerParams.SliceAxis.Name(params.axis)
    if axis == "CHANNEL_AXIS":
        N = C
    if axis == "HEIGHT_AXIS":
        N = H
    if axis == "WIDTH_AXIS":
        N = W
    if end < 0:
        end = end + N
    end = min(end, N)
    if start > N - 1:
        L = 0
    else:
        L = np.floor((end - 1 - start) / stride) + 1
        if L < 0:
            L = 0
    if axis == "CHANNEL_AXIS":
        C = L
    if axis == "HEIGHT_AXIS":
        H = L
    if axis == "WIDTH_AXIS":
        W = L
    shape_dict[layer.output[0]] = (Seq, Batch, int(C), int(H), int(W))


def _simple_recurrent(layer, shape_dict):
    params = layer.simpleRecurrent
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]

    Cout = params.outputVectorSize
    if params.sequenceOutput:
        shape_dict[layer.output[0]] = (Seq, Batch, int(Cout), 1, 1)
    else:
        shape_dict[layer.output[0]] = (1, Batch, int(Cout), 1, 1)
    shape_dict[layer.output[1]] = (1, Batch, int(Cout), 1, 1)


def _gru(layer, shape_dict):
    params = layer.gru
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]

    Cout = params.outputVectorSize
    if params.sequenceOutput:
        shape_dict[layer.output[0]] = (Seq, Batch, int(Cout), 1, 1)
    else:
        shape_dict[layer.output[0]] = (1, Batch, int(Cout), 1, 1)
    shape_dict[layer.output[1]] = (1, Batch, int(Cout), 1, 1)


def _uni_directional_lstm(layer, shape_dict):
    params = layer.uniDirectionalLSTM
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]

    Cout = params.outputVectorSize
    if params.params.sequenceOutput:
        shape_dict[layer.output[0]] = (Seq, Batch, int(Cout), 1, 1)
    else:
        shape_dict[layer.output[0]] = (1, Batch, int(Cout), 1, 1)
    shape_dict[layer.output[1]] = (1, Batch, int(Cout), 1, 1)
    shape_dict[layer.output[2]] = (1, Batch, int(Cout), 1, 1)


def _bi_directional_lstm(layer, shape_dict):
    params = layer.biDirectionalLSTM
    Seq, Batch, C, H, W = shape_dict[layer.input[0]]
    Cout = params.outputVectorSize
    if params.params.sequenceOutput:
        shape_dict[layer.output[0]] = (Seq, Batch, 2 * int(Cout), 1, 1)
    else:
        shape_dict[layer.output[0]] = (1, Batch, 2 * int(Cout), 1, 1)
    shape_dict[layer.output[1]] = (1, Batch, int(Cout), 1, 1)
    shape_dict[layer.output[2]] = (1, Batch, int(Cout), 1, 1)
    shape_dict[layer.output[3]] = (1, Batch, int(Cout), 1, 1)
    shape_dict[layer.output[4]] = (1, Batch, int(Cout), 1, 1)


_LAYER_REGISTERY = {
    "convolution": _convolution,
    "pooling": _pooling,
    "activation": _identity,
    "innerProduct": _inner_product,
    "embedding": _embedding,
    "batchnorm": _identity,
    "mvn": _identity,
    "l2normalize": _identity,
    "softmax": _identity,
    "lrn": _identity,
    "crop": _crop,
    "padding": _padding,
    "upsample": _upsample,
    "unary": _identity,
    "add": _add,
    "multiply": _add,
    "average": _add,
    "scale": _add,
    "bias": _add,
    "max": _add,
    "min": _add,
    "dot": _dot,
    "reduce": _reduce,
    "loadConstant": _load_constant,
    "reshape": _reshape,
    "flatten": _flatten,
    "permute": _permute,
    "concat": _concat,
    "split": _split,
    "sequenceRepeat": _sequence_repeat,
    "reorganizeData": _reorganize_data,
    "slice": _slice,
    "simpleRecurrent": _simple_recurrent,
    "gru": _gru,
    "uniDirectionalLSTM": _uni_directional_lstm,
    "biDirectionalLSTM": _bi_directional_lstm,
}


def infer_shapes(nn_spec, input_spec, input_shape_dict=None):
    """
    Input:

        spec : mlmodel spec
        input_shape_dict: dictionary of  string --> tuple
                      string:  input name
                      tuple: input shape as a 5 length tuple in order (Seq, Batch, C, H, W)

        If input_shape_dict is not provided, input shapes are inferred from the input description in the mlmodel.
        Since the description in the specification only contains values of C,H,W; Seq and Batch dimensions are set to 1.

    Output:

        shape_dict:  dictionary containing all the blobs in the neural network and their shapes, expressed as length 5 tuples,
                     to be interpreted in order (Seq, Batch, C, H, W).
    """

    shape_dict = {}
    if input_shape_dict:
        for key, value in input_shape_dict.items():
            assert len(value) == 5, "Shape of the input must be of length 5"
            shape_dict[key] = value

    # construct input_shape_dict from the model description
    else:
        for inp in input_spec:
            input_name = inp.name
            C = H = W = 1
            if inp.type.WhichOneof("Type") == "imageType":
                W = int(inp.type.imageType.width)
                H = int(inp.type.imageType.height)
                colorspace = _FeatureTypes_pb2.ImageFeatureType.ColorSpace.Name(
                    inp.type.imageType.colorSpace
                )
                if colorspace == "GRAYSCALE":
                    C = 1
                elif colorspace == "RGB" or colorspace == "BGR":
                    C = 3
                else:
                    raise ValueError("Input %s : Invalid Colorspace" % (input_name))
            elif inp.type.WhichOneof("Type") == "multiArrayType":
                array_shape = inp.type.multiArrayType.shape
                if len(array_shape) == 1:
                    C = array_shape[0]
                elif len(array_shape) == 3:
                    C, H, W = map(int, array_shape)
                else:
                    raise ValueError(
                        "Input %s : Multi array must be of length 1 or 3" % (input_name)
                    )
            else:
                raise ValueError(
                    "Input %s : Input type must be image or multi-array" % (input_name)
                )
            shape_dict[input_name] = (1, 1, C, H, W)

    layers = nn_spec.layers

    for i, layer in enumerate(layers):
        for inp in layer.input:
            assert inp in shape_dict, "Input %s shape not cannot be determined" % (inp)
        layer_type = layer.WhichOneof("layer")
        if layer_type == "custom":
            break
        layer_translator = _get_translator_function(layer_type)
        layer_translator(layer, shape_dict)

    return shape_dict



from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.input_types import ImageType, Shape, EnumeratedShapes

from coremltools.converters.mil.mil import Builder as mb


@register_pass(namespace='common')
def image_input_preprocess(prog):
    """
    Plug in transpose for image input that were NHWC to NCHW.

    Steps:
    1. Check whether there are any inputs that the users specify is ImageType.
    2. Check the channel's dimension for all ImageType
        a) channel_first == True
            We do not modify this input, since channel_first is the intended
            behaviour for feeding images for optimal performance
        b) channel_first == False
            We convert the input into a "channel_first" input, and plug in a 
            transpose for the input to maintain the remaining graph's dimensionality.
    """
    for f_name, f in prog.functions.items():
        if f_name == 'main':
            # We need to make sure main exist and start here.
            _image_input_preprocess(prog)

def _transform_to_channel_first(shape):
    if isinstance(shape, tuple):
        shape = list(shape)
        return tuple(shape[:-3] + [shape[-1]] + shape[-3:-1])
    else:
        return shape[:-3] + [shape[-1]] + shape[-3:-1]

def _image_input_preprocess(prog):
    main_input_types = list(prog.main_input_types)
    for idx, input_type in enumerate(main_input_types):
        if isinstance(input_type, ImageType) and not input_type.channel_first:
            name = input_type.name
            # Build new ImageType to change data layout
            if isinstance(input_type.shape, Shape):
                new_shape = _transform_to_channel_first(input_type.shape.shape)
                new_default = _transform_to_channel_first(input_type.shape.default)
                shape_type = Shape(shape=new_shape, default=new_default)
            elif isinstance(input_type.shape, EnumeratedShapes):
                shape_list = []
                for shape in input_type.shape.shapes:
                    if isinstance(shape, Shape):
                        shape_list.append(_transform_to_channel_first(shape.shape))
                    else:
                        shape_list.append(_transform_to_channel_first(shape))
                shape_type = EnumeratedShapes(shapes=shape_list,
                                              default=_transform_to_channel_first(input_type.shape.default))
            new_image_type = ImageType(name=name,
                                       shape=shape_type,
                                       bias=input_type.bias,
                                       scale=input_type.scale,
                                       color_layout=input_type.color_layout,
                                       channel_first=True)
            main_input_types[idx] = new_image_type

            # Reconstruct Placeholder of Function inputs.
            placeholder_op = prog.functions['main'].placeholder_inputs[name]
            old_var = placeholder_op.outputs[0]
            nchw_shape = _transform_to_channel_first(placeholder_op.sym_shape)
            placeholder_op.__init__(nchw_shape, dtype=placeholder_op.dtype, name=placeholder_op.name)

            # Update Function input var
            prog.functions['main']._input_dict[name] = placeholder_op.outputs[0]
            prog.functions['main'].function_inputs = tuple(prog.functions['main']._input_dict.values())

            # Add transpose into graph (Transpose from NCHW back to NHWC)
            curr_block = prog.functions['main']
            curr_var = prog.functions['main'].inputs[name]

            perm = list(range(curr_var.rank))
            perm = perm[:-3] + [perm[-2], perm[-1], perm[-3]]
            with curr_block:
                new_input = mb.transpose(x=curr_var,
                                         perm=perm,
                                         before_op=prog.functions['main'].operations[0],
                                         name=curr_var.name+"__transpose_from_nchw__")
            curr_block.replace_uses_of_var_after_op(anchor_op=None,
                                                    old_var=old_var,
                                                    new_var=new_input)
    prog.main_input_types = tuple(main_input_types)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *
from coremltools.converters.mil.testing_utils import get_core_ml_prediction
from coremltools._deps import _IS_MACOS
from .testing_utils import UNK_SYM, run_compare_builder

backends = testing_reqs.backends


class TestRandomBernoulli:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):

        x_val = np.array([0.0], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                mb.add(x=x, y=x),
                mb.random_bernoulli(shape=np.array([2, 1, 3], np.int32), prob=1.0),
                mb.random_bernoulli(shape=np.array([3, 1, 2], np.int32), prob=0.0),
            ]

        expected_outputs = [
            np.array(np.zeros(shape=(1,)), np.float32),
            np.array(np.ones(shape=(2, 1, 3)), np.float32),
            np.array(np.zeros(shape=(3, 1, 2)), np.float32),
        ]

        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, rank, prob, dynamic",
        itertools.product(
            [True, False],
            backends,
            [rank for rank in range(1, 6)],
            [1.0, 0.0],
            [True, False],
        ),
    )
    def test_builder_to_backend_stress(
        self, use_cpu_only, backend, rank, prob, dynamic
    ):
        shape = np.random.randint(low=1, high=4, size=rank).astype(np.int32)
        x_val = np.array([0.0], dtype=np.float32)
        if dynamic:
            input_placeholders = {
                "x": mb.placeholder(shape=x_val.shape),
                "dyn_shape": mb.placeholder(shape=shape.shape, dtype=types.int32),
            }
            input_values = {"x": x_val, "dyn_shape": shape}
        else:
            input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
            input_values = {"x": x_val}

        def build(x):
            return [mb.add(x=x, y=x), mb.random_bernoulli(shape=shape, prob=prob)]

        def build_dyn(x, dyn_shape):
            return [mb.add(x=x, y=x), mb.random_bernoulli(shape=dyn_shape, prob=prob)]

        expected_outputs = [
            np.array(np.zeros(shape=(1,)), np.float32),
            np.random.binomial(1, prob, shape),
        ]

        if dynamic:
            expected_output_types = [
                tuple([UNK_SYM for _ in o.shape]) + (types.fp32,)
                for o in expected_outputs
            ]
        else:
            expected_output_types = [
                o.shape[:] + (types.fp32,) for o in expected_outputs
            ]

        builder = build_dyn if dynamic else build

        run_compare_builder(
            builder,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestRandomCategorical:
    def softmax(self, data):
        e_data = np.exp(data - np.max(data))
        return e_data / e_data.sum()

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array([1], dtype=np.int32)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                mb.random_categorical(x=x, seed=1),
                mb.random_categorical(x=x, seed=1, size=4),
            ]

        expected_outputs = [
            np.array(np.zeros(shape=(1,)), dtype=np.float32),
            np.array(np.zeros(shape=(4,)), dtype=np.float32),
        ]

        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, n_sample, n_class",
        itertools.product([True, False], backends, [50000], [2, 10, 20]),
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, n_sample, n_class):
        output_name = "random_categorical"
        logits = np.random.rand(2, n_class)
        probs = [self.softmax(logits[0]), self.softmax(logits[1])]

        # Test logits input
        input_placeholders = {"x": mb.placeholder(shape=(2, n_class))}
        input_values = {"x": logits}

        def build(x):
            return [
                mb.random_categorical(
                    x=x, size=n_sample, mode="logits", name=output_name
                )
            ]

        if _IS_MACOS:
            prediction = get_core_ml_prediction(
                build, input_placeholders, input_values, backend=backend
            )

            ref0 = np.random.multinomial(n_sample, probs[0])
            ref1 = np.random.multinomial(n_sample, probs[1])

            pred0 = prediction[output_name].reshape(2, n_sample)[0]
            pred1 = prediction[output_name].reshape(2, n_sample)[1]

            # convert to bincount and validate probabilities
            pred0 = np.bincount(np.array(pred0).astype(np.int), minlength=n_class)
            pred1 = np.bincount(np.array(pred1).astype(np.int), minlength=n_class)

            assert np.allclose(np.true_divide(pred0, n_sample), probs[0], atol=1e-2)
            assert np.allclose(
                np.true_divide(pred0, n_sample),
                np.true_divide(ref0, n_sample),
                atol=1e-2,
            )

            assert np.allclose(np.true_divide(pred1, n_sample), probs[1], atol=1e-2)
            assert np.allclose(
                np.true_divide(pred1, n_sample),
                np.true_divide(ref1, n_sample),
                atol=1e-2,
            )

        # Test probs input
        input_placeholders = {"x": mb.placeholder(shape=(2, n_class))}
        input_values = {"x": np.array(probs)}

        def build(x):
            return [
                mb.random_categorical(
                    x=x, size=n_sample, mode="probs", name=output_name
                )
            ]

        if _IS_MACOS:
            prediction = get_core_ml_prediction(
                build, input_placeholders, input_values, backend=backend
            )

            pred0 = prediction[output_name].reshape(2, n_sample)[0]
            pred1 = prediction[output_name].reshape(2, n_sample)[1]

            # convert to bincount and validate probabilities
            pred0 = np.bincount(np.array(pred0).astype(np.int), minlength=n_class)
            pred1 = np.bincount(np.array(pred1).astype(np.int), minlength=n_class)

            assert np.allclose(np.true_divide(pred0, n_sample), probs[0], atol=1e-2)
            assert np.allclose(
                np.true_divide(pred0, n_sample),
                np.true_divide(ref0, n_sample),
                atol=1e-2,
            )

            assert np.allclose(np.true_divide(pred1, n_sample), probs[1], atol=1e-2)
            assert np.allclose(
                np.true_divide(pred1, n_sample),
                np.true_divide(ref1, n_sample),
                atol=1e-2,
            )


class TestRandomNormal:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array([0.0], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                mb.add(x=x, y=x),
                mb.random_normal(
                    shape=np.array([2, 1, 3], np.int32), mean=1.0, stddev=0.0
                ),
                mb.random_normal(
                    shape=np.array([3, 1, 2], np.int32), mean=0.0, stddev=0.0
                ),
            ]

        expected_outputs = [
            np.array(np.zeros(shape=(1,)), np.float32),
            np.array(np.ones(shape=(2, 1, 3)), np.float32),
            np.array(np.zeros(shape=(3, 1, 2)), np.float32),
        ]

        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, rank, mean, dynamic",
        itertools.product(
            [True, False],
            backends,
            [rank for rank in range(1, 6)],
            [1.0, 0.0],
            [True, False],
        ),
    )
    def test_builder_to_backend_stress(
        self, use_cpu_only, backend, rank, mean, dynamic
    ):
        shape = np.random.randint(low=1, high=4, size=rank).astype(np.int32)
        x_val = np.array([0.0], dtype=np.float32)
        if dynamic:
            input_placeholders = {
                "x": mb.placeholder(shape=x_val.shape),
                "dyn_shape": mb.placeholder(shape=shape.shape, dtype=types.int32),
            }
            input_values = {"x": x_val, "dyn_shape": shape}
        else:
            input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
            input_values = {"x": x_val}

        def build(x):
            return [
                mb.add(x=x, y=x),
                mb.random_normal(shape=shape, mean=mean, stddev=0.0),
            ]

        def build_dyn(x, dyn_shape):
            return [
                mb.add(x=x, y=x),
                mb.random_normal(shape=dyn_shape, mean=mean, stddev=0.0),
            ]

        expected_outputs = [
            np.array(np.zeros(shape=(1,)), np.float32),
            np.random.normal(loc=mean, scale=0.0, size=shape),
        ]

        if dynamic:
            expected_output_types = [
                tuple([UNK_SYM for _ in o.shape]) + (types.fp32,)
                for o in expected_outputs
            ]
        else:
            expected_output_types = [
                o.shape[:] + (types.fp32,) for o in expected_outputs
            ]

        builder = build_dyn if dynamic else build
        run_compare_builder(
            builder,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestRandomUniform:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array([0.0], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                mb.add(x=x, y=x),
                mb.random_uniform(
                    shape=np.array([2, 1, 3], np.int32), low=0.0, high=0.0
                ),
                mb.random_uniform(
                    shape=np.array([3, 1, 2], np.int32), low=1.0, high=1.0
                ),
            ]

        expected_outputs = [
            np.array(np.zeros(shape=(1,)), np.float32),
            np.array(np.zeros(shape=(2, 1, 3)), np.float32),
            np.array(np.ones(shape=(3, 1, 2)), np.float32),
        ]

        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, rank, low, high, dynamic",
        itertools.product(
            [True, False],
            backends,
            [rank for rank in range(1, 6)],
            [0.0],
            [0.0],
            [True, False],
        ),
    )
    def test_builder_to_backend_stress(
        self, use_cpu_only, backend, rank, low, high, dynamic
    ):
        shape = np.random.randint(low=1, high=4, size=rank).astype(np.int32)
        x_val = np.array([0.0], dtype=np.float32)
        if dynamic:
            input_placeholders = {
                "x": mb.placeholder(shape=x_val.shape),
                "dyn_shape": mb.placeholder(shape=shape.shape, dtype=types.int32),
            }
            input_values = {"x": x_val, "dyn_shape": shape}
        else:
            input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
            input_values = {"x": x_val}

        def build(x):
            return [
                mb.add(x=x, y=x),
                mb.random_uniform(shape=shape, low=low, high=high),
            ]

        def build_dyn(x, dyn_shape):
            return [
                mb.add(x=x, y=x),
                mb.random_uniform(shape=dyn_shape, low=low, high=high),
            ]

        expected_outputs = [
            np.array(np.zeros(shape=(1,)), np.float32),
            np.random.uniform(low=low, high=high, size=shape),
        ]

        if dynamic:
            expected_output_types = [
                tuple([UNK_SYM for _ in o.shape]) + (types.fp32,)
                for o in expected_outputs
            ]
        else:
            expected_output_types = [
                o.shape[:] + (types.fp32,) for o in expected_outputs
            ]

        builder = build_dyn if dynamic else build
        run_compare_builder(
            builder,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs=expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools._deps import _HAS_TORCH

register_torch_op = None

if _HAS_TORCH:
    from .load import load
    from .torch_op_registry import register_torch_op



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import itertools
import pandas as pd
import os
import unittest

from coremltools._deps import _HAS_SKLEARN
from coremltools.converters.sklearn import convert
from coremltools.models.utils import (
    evaluate_classifier,
    evaluate_classifier_with_probabilities,
    _macos_version,
    _is_macos,
)

if _HAS_SKLEARN:
    from sklearn.linear_model import LogisticRegression
    from sklearn.svm import LinearSVC


@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
class GlmCassifierTest(unittest.TestCase):
    def test_logistic_regression_binary_classification_with_string_labels(self):
        self._conversion_and_evaluation_helper_for_logistic_regression(["Foo", "Bar"])

    def test_logistic_regression_multiclass_classification_with_int_labels(self):
        self._conversion_and_evaluation_helper_for_logistic_regression([1, 2, 3, 4])

    @staticmethod
    def _generate_random_data(labels):
        import random

        random.seed(42)

        # Generate some random data
        x, y = [], []
        for _ in range(100):
            x.append([random.gauss(2, 3), random.gauss(-1, 2)])
            y.append(random.choice(labels))
        return x, y

    def _conversion_and_evaluation_helper_for_logistic_regression(self, class_labels):
        options = {
            "C": (0.1, 1.0, 2.0),
            "fit_intercept": (True, False),
            "class_weight": ("balanced", None),
            "solver": ("newton-cg", "lbfgs", "liblinear", "sag"),
        }

        # Generate a list of all combinations of options and the default parameters
        product = itertools.product(*options.values())
        args = [{}] + [dict(zip(options.keys(), p)) for p in product]

        x, y = GlmCassifierTest._generate_random_data(class_labels)
        column_names = ["x1", "x2"]
        df = pd.DataFrame(x, columns=column_names)

        for cur_args in args:
            print(class_labels, cur_args)
            cur_model = LogisticRegression(**cur_args)
            cur_model.fit(x, y)

            spec = convert(
                cur_model, input_features=column_names, output_feature_names="target"
            )

            if _is_macos() and _macos_version() >= (10, 13):
                probability_lists = cur_model.predict_proba(x)
                df["classProbability"] = [
                    dict(zip(cur_model.classes_, cur_vals))
                    for cur_vals in probability_lists
                ]

                metrics = evaluate_classifier_with_probabilities(
                    spec, df, probabilities="classProbability", verbose=False
                )
                self.assertEquals(metrics["num_key_mismatch"], 0)
                self.assertLess(metrics["max_probability_error"], 0.00001)

    def test_linear_svc_binary_classification_with_string_labels(self):
        self._conversion_and_evaluation_helper_for_linear_svc(["Foo", "Bar"])

    def test_linear_svc_multiclass_classification_with_int_labels(self):
        self._conversion_and_evaluation_helper_for_linear_svc([1, 2, 3, 4])

    def _conversion_and_evaluation_helper_for_linear_svc(self, class_labels):
        ARGS = [
            {},
            {"C": 0.75, "loss": "hinge"},
            {"penalty": "l1", "dual": False},
            {"tol": 0.001, "fit_intercept": False},
            {"intercept_scaling": 1.5},
        ]

        x, y = GlmCassifierTest._generate_random_data(class_labels)
        column_names = ["x1", "x2"]
        df = pd.DataFrame(x, columns=column_names)

        for cur_args in ARGS:
            print(class_labels, cur_args)
            cur_model = LinearSVC(**cur_args)
            cur_model.fit(x, y)

            spec = convert(
                cur_model, input_features=column_names, output_feature_names="target"
            )

            if _is_macos() and _macos_version() >= (10, 13):
                df["prediction"] = cur_model.predict(x)

                cur_eval_metics = evaluate_classifier(spec, df, verbose=False)
                self.assertEquals(cur_eval_metics["num_errors"], 0)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _
from ..basic_graph_ops import delete_node
import logging
import sys

def delete_fakequant_node_and_repair_graph(g, node):
    inputs = node.inputs
    # Delete const inputs of the fakequant op
    for i in inputs:
        if g[i].op == 'Const':
            delete_node(g, i)
        else:
            non_const_input = i
    outputs = node.outputs
    # Append FakeQuant Op's outputs to its input node's outputs
    g[non_const_input].outputs = [i for i in g[non_const_input].outputs if i != node.name]
    g[non_const_input].outputs.extend(outputs)
    # Modify the FakeQuant op's outputs to set FakeQuant op's parent node as the new input.
    for i in outputs:
        for j in range(len(g[i].inputs)):
            if g[i].inputs[j] == node.name:
                g[i].inputs[j] = non_const_input
    delete_node(g, node)

def quantization_pass_impl(fn):
    all_quantization_ops = [i for i in fn.graph.values() if "FakeQuant" in i.op]
    for node in all_quantization_ops:
        is_const_input = True
        for input in node.inputs:
            if fn.graph[input].op != 'Const':
                is_const_input = False
        if not is_const_input and ('weights_quant' not in input):
            # If activation quantization -
            # Delete the FakeQuant op and its const inputs,
            # Append FakeQuant Op's outputs to its input node's outputs,
            # Modify the FakeQuant op's outputs to reflect the 'new' input node.
            delete_fakequant_node_and_repair_graph(fn.graph, node)
        else:
            # If weight quantization -
            # Add attributes of the FakeQuant op to its output's attr dict
            for output in node.outputs:
                output_node = fn.graph[output]
                output_node.attr['quantize'] = True
                output_node.attr['num_bits'] = node.attr['num_bits']
                output_node.attr['narrow_range'] = node.attr['narrow_range']
                output_node.attr['quantize_min'] = fn.graph[node.inputs[1]].value.val
                output_node.attr['quantize_max'] = fn.graph[node.inputs[2]].value.val

def quantization_pass(tfssa):
    """
    Delete activation quantization ops and repair TF graph:
        If the FakeQuant op is not connected to constant inputs (which means that the op performs activation
        quantization) then delete that FakeQuant op and repair the graph.
    Edit weight quantization ops:
        If the FakeQuant op is connected to constant inputs then add its attributes to its output op so that parameters
        min, max, narrow_range, num_bits are available (in addition to weights) to downstream ops for denoting and
        supporting weight quantization.
    """
    for v in tfssa.functions.values():
        quantization_pass_impl(v)


# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from ._tree_ensemble import convert_tree_ensemble as _convert_tree_ensemble
from ...models import MLModel as _MLModel
from coremltools import __version__ as ct_version
from coremltools.models import _METADATA_VERSION, _METADATA_SOURCE


def convert(
    model,
    feature_names=None,
    target="target",
    force_32bit_float=True,
    mode="regressor",
    class_labels=None,
    n_classes=None,
):
    """
    Convert a trained XGBoost model to Core ML format.

    Parameters
    ----------
    decision_tree : Booster
        A trained XGboost tree model.

    feature_names: [str] | str
        Names of input features that will be exposed in the Core ML model
        interface.

        Can be set to one of the following:

        - None for using the feature names from the model.
        - List of names of the input features that should be exposed in the
          interface to the Core ML model. These input features are in the same
          order as the XGboost model.

    target: str
        Name of the output feature name exposed to the Core ML model.

    force_32bit_float: bool
        If True, then the resulting CoreML model will use 32 bit floats internally.

    mode: str in ['regressor', 'classifier']
        Mode of the tree model.

    class_labels: list[int] or None
        List of classes. When set to None, the class labels are just the range from
        0 to n_classes - 1.

    n_classes: int or None
        Number of classes in classification. When set to None, the number of
        classes is expected from the model or class_labels should be provided.

    Returns
    -------
    model:MLModel
        Returns an MLModel instance representing a Core ML model.

    Examples
    --------
    .. sourcecode:: python

		# Convert it with default input and output names
   		>>> import coremltools
		>>> coreml_model = coremltools.converters.xgboost.convert(model)

		# Saving the Core ML model to a file.
		>>> coremltools.save('my_model.mlmodel')
    """
    model = _MLModel(
        _convert_tree_ensemble(
            model,
            feature_names,
            target,
            force_32bit_float=force_32bit_float,
            mode=mode,
            class_labels=class_labels,
            n_classes=n_classes,
        )
    )

    from xgboost import __version__ as xgboost_version

    model.user_defined_metadata[_METADATA_VERSION] = ct_version
    model.user_defined_metadata[_METADATA_SOURCE] = "xgboost=={0}".format(
        xgboost_version
    )

    return model
